#define NDEBUG

#include "algorithm.h"

#include <iostream>
#include <sstream>
#include <stdexcept>

using namespace std;

class solve_progress_handler: public solve_progress_listener {
public:
    solve_progress_handler();
private:
    bool _started;
    void node_visited(span entry_array[], entry_count_t entry_count);
};

static string
format_solution(const span entry_array[], entry_count_t entry_count)
{
    stringstream stream;
    stream << "{0";
    for (
        entry_count_t entry_index = 0;
        entry_index < entry_count - 1;
        ++entry_index)
        stream << ", " << entry_array[entry_index];
    stream << "}";
    return stream.str();
}

static bool is_cmd_option(const char *arg, const string &ref_cmd_option) {
    if (arg[0] == '-' || arg[0] == '/') {
        string::size_type size = ref_cmd_option.size();
        if (ref_cmd_option.compare(0, size, arg + 1, size) == 0) return true;
    }
    return false;
}

static bool
is_terminated_cmd_option(
const char *arg,
const string &ref_cmd_option,
char terminator = '\0')
{
    return
        is_cmd_option(arg, ref_cmd_option) &&
        arg[ref_cmd_option.size() + 1] == terminator;
}

static void test_unique_cmd_option(const string &ref_cmd_option, bool used) {
    if (used)
        cerr << "duplicated option \"-" << ref_cmd_option << "\"" << endl <<
        endl;
}

static bool
parse_predicated_cmd_option(
const char *arg,
const string &ref_cmd_option,
unsigned int &ref_predicate)
{
    test_unique_cmd_option(ref_cmd_option, ref_predicate != 0);
    stringstream(arg + (ref_cmd_option.size() + 2)) >> ref_predicate;
    if (ref_predicate == 0) {
        cerr <<
            "invalid argument for \"-" << ref_cmd_option << "\"" << endl <<
            endl;
        return false;
    }
    return true;
}

static bool
parse_boolean_cmd_option(const string &ref_cmd_option, bool &ref_value)
{
    test_unique_cmd_option(ref_cmd_option, ref_value);
    ref_value = true;
    return true;
}

static bool parse_args(
int argc,
char *argv[],
unsigned int &ref_max_solution_count,
unsigned int &ref_max_visited_node_count,
bool &ref_verbose)
{
    const string cmd_option_help("?");
    for (int arg_index = 1; arg_index < argc; ++arg_index) {
        char *arg = argv[arg_index];
        if (is_cmd_option(arg, cmd_option_help)) {
            cout <<
                "Turnpike 1.0 - Copyright 2008 Francesco Trotta - www.ft1.org"
                << endl << endl <<
                "Solves the turnpike problem." << endl << endl;
            return false;
        }
    }
    unsigned int max_solution_count = 0;
    unsigned int max_visited_node_count = 0;
    bool verbose = false;
    const string cmd_option_maxsols("maxsols");
    const string cmd_option_maxnodes("maxnodes");
    const string cmd_option_verbose("verbose");
    for (int arg_index = 1; arg_index < argc; ++arg_index) {
        char *arg = argv[arg_index];
        if (is_terminated_cmd_option(arg, cmd_option_maxsols, ':')) {
            if (
                !parse_predicated_cmd_option(
                arg,
                cmd_option_maxsols,
                max_solution_count))
                return false;
        }
        else if (is_terminated_cmd_option(arg, cmd_option_maxnodes, ':')) {
            if (
                !parse_predicated_cmd_option(
                arg,
                cmd_option_maxnodes,
                max_visited_node_count))
                return false;
        }
        else if (is_terminated_cmd_option(arg, cmd_option_verbose)) {
            if (!parse_boolean_cmd_option(cmd_option_verbose, verbose))
                return false;
        }
        else {
            cerr << "syntax error" << endl << endl;
            return false;
        }
    }
    ref_max_solution_count =
        max_solution_count == 0 ? UINT_MAX : max_solution_count;
    ref_max_visited_node_count =
        max_visited_node_count == 0 ? UINT_MAX : max_visited_node_count;
    ref_verbose = verbose;
    return true;
}

static bool parse_in_val(const string in_val_str, span &ref_in_val) {
    // Under VC++ stringstream::operator >>(_ULONGLONG &) only parses numbers up
    // to 9999999999999999999, but unsigned long long integers may range up to
    // 18446744073709551615; that's why we do not use the built-in
    // functionality.
    span in_val = 0;
    for (string::size_type index = 0; index < in_val_str.length(); ++index) {
        if (in_val > (~static_cast<span>(0)) / 10) return false;

        unsigned int digit = in_val_str.c_str()[index] - '0';
        if (digit >= 10) return false;

        in_val = in_val * 10 + digit;
        if (in_val < digit) return false; // overflow!
    }
    ref_in_val = in_val;
    return true;
}

static multiset<span> get_prob_data() {
    multiset<span> in_val_multiset;
    string prob_data;
    do {
        cout << "Please, enter problem data:\n";
        cin.clear(ios::goodbit);
        getline(cin, prob_data);
    } while (prob_data.empty());
    static const char separator[] = " \n\r\t\v,;";
    string::size_type end_pos = 0;
    for (;;) {
        string::size_type offset =
            prob_data.find_first_not_of(separator, end_pos);
        if (offset == string::npos) break;
        end_pos = prob_data.find_first_of(separator, offset);
        string in_val_str =
            prob_data.substr(
            offset,
            end_pos != string::npos ? end_pos - offset : string::npos
            );
        span in_val;
        if (!parse_in_val(in_val_str, in_val))
            throw runtime_error("input values invalid or too large");
        in_val_multiset.insert(in_val);
    }
    return in_val_multiset;
}

static int sub_main(int argc, char *argv[]) {

    // parse command line arguments
    unsigned int max_solution_count;
    unsigned int max_visited_node_count;
    bool verbose;
    if (
        !parse_args(
        argc,
        argv,
        max_solution_count,
        max_visited_node_count,
        verbose))
    {
        // parsing error or flag "-?" found
        cout <<
            "Usage:\tTURNPIKE [-maxsols:<n>] [-maxnodes:<n>]\n\n"
            "  -maxsols:<n>     find at most <n> solutions\n"
            "  -maxnodes:<n>    visit at most <n> nodes\n"
            "  -verbose         output nodes as they are visited\n\n"
            "Use input redirection syntax to read the problem data from a file."
            "\n"
            "Input values should be separated by spaces, commas, colons, or a "
            "combination of them.\n"
            "If \"-verbose\" is specified, the computation time measurement "
            "can be inaccurate.\n\n"
            "Examples:\n\n"
            "\tTURNPIKE\n"
            "\tTURNPIKE -maxsols:2 -maxnodes:100 <input.txt >output.txt\n";
        return 1;
    }

    // get problem data
    multiset<span> in_val_multiset = get_prob_data();

    // calculate solution
    unsigned int visited_node_count;
    solution_set best_set;
    unsigned long long elapsed_time;
    if (verbose) {
        solve_progress_handler handler;
        visited_node_count =
            solve(
            in_val_multiset,
            max_solution_count,
            max_visited_node_count,
            &handler,
            best_set,
            elapsed_time
            );
        cout << endl;
    }
    else {
        visited_node_count =
            solve(
            in_val_multiset,
            max_solution_count,
            max_visited_node_count,
            NULL,
            best_set,
            elapsed_time
            );
    }

    // print solution and computation data
    if (best_set.size() > 0) {
        cout << "Found " << best_set.size() <<
            (best_set.size() == 1 ? " solution" : " solutions") << ":" << endl;
        for (
            solution_set::iterator iterator = best_set.begin();
            iterator != best_set.end();
            ++iterator)
        {
            const solution &ref_best = *iterator;
            cout << endl <<
                format_solution(ref_best, best_set.get_entry_count());
        }
    }
    else
        cout << "No solutions found.";
    cout <<
        endl <<
        endl << "Visited nodes:    " << visited_node_count <<
        endl << "Computation time: ";
    if (elapsed_time > 0)
        cout << elapsed_time << " microsecs";
    else
        cout << "-";
    cout << endl;
    return 0;
}

int __cdecl main(int argc, char *argv[]) {
    try {
        return sub_main(argc, argv);
    }
    catch (bad_alloc exc) {
        cerr << exc.what() << endl;
    }
    catch (invalid_argument exc) {
        cerr << exc.what() << endl;
    }
    catch (runtime_error exc) {
        cerr << exc.what() << endl;
    }
    return 1;
}

solve_progress_handler::solve_progress_handler(): _started(false) {
}

void
solve_progress_handler::node_visited(
span entry_array[],
entry_count_t entry_count)
{
    if (!_started) {
        cout << "Listing visited nodes:" << endl << endl;
        _started = true;
    }
    cout << format_solution(entry_array, entry_count) << endl;
}
