#pragma once

#include <assert.h>
#include <stddef.h>
#include <iterator>

template<typename T> class basic_auto_array;

template<typename T>
class basic_auto_array_iterator {
    template<typename U> friend class basic_auto_array;
public:
    typedef ptrdiff_t difference_type;
    typedef std::random_access_iterator_tag iterator_category;
    typedef T *pointer;
    typedef T &reference;
    typedef T value_type;
    basic_auto_array_iterator operator ++(int);
    basic_auto_array_iterator operator --(int);
    basic_auto_array_iterator &operator ++();
    basic_auto_array_iterator &operator --();
    T &operator *() const;
    basic_auto_array_iterator operator +(difference_type offset) const;
    basic_auto_array_iterator operator -(difference_type offset) const;
    difference_type operator -(const basic_auto_array_iterator &arg) const;
    bool operator <(const basic_auto_array_iterator &arg) const;
    bool operator >(const basic_auto_array_iterator &arg) const;
    bool operator <=(const basic_auto_array_iterator &arg) const;
    bool operator >=(const basic_auto_array_iterator &arg) const;
    bool operator ==(const basic_auto_array_iterator &arg) const;
    bool operator !=(const basic_auto_array_iterator &arg) const;
private:
    T *_ptr;
    basic_auto_array_iterator(); // no default constructor
    basic_auto_array_iterator(T *ptr);
};

template<typename T>
basic_auto_array_iterator<T>::basic_auto_array_iterator(T *ptr): _ptr(ptr) {
}

template<typename T>
basic_auto_array_iterator<T> basic_auto_array_iterator<T>::operator ++(int) {
    assert(_ptr != NULL);
    return basic_auto_array_iterator(_ptr++);
}

template<typename T>
basic_auto_array_iterator<T> basic_auto_array_iterator<T>::operator --(int) {
    assert(_ptr != NULL);
    return basic_auto_array_iterator(_ptr--);
}

template<typename T>
basic_auto_array_iterator<T> &basic_auto_array_iterator<T>::operator ++() {
    assert(_ptr != NULL);
    ++_ptr;
    return *this;
}

template<typename T>
basic_auto_array_iterator<T> &basic_auto_array_iterator<T>::operator --() {
    assert(_ptr != NULL);
    --_ptr;
    return *this;
}

template<typename T>
T &basic_auto_array_iterator<T>::operator *() const {
    assert(_ptr != NULL);
    return *_ptr;
}

template<typename T>
basic_auto_array_iterator<T>
basic_auto_array_iterator<T>::operator +(difference_type offset)
const
{
    return basic_auto_array_iterator(_ptr + offset);
}

template<typename T>
basic_auto_array_iterator<T>
basic_auto_array_iterator<T>::operator -(difference_type offset)
const
{
    return basic_auto_array_iterator(_ptr - offset);
}

template<typename T>
typename basic_auto_array_iterator<T>::difference_type
basic_auto_array_iterator<T>::operator -(const basic_auto_array_iterator &arg)
const
{
    return _ptr - arg._ptr;
}

template<typename T>
bool
basic_auto_array_iterator<T>::operator <(const basic_auto_array_iterator &arg)
const
{
    return _ptr < arg._ptr;
}

template<typename T>
bool
basic_auto_array_iterator<T>::operator >(const basic_auto_array_iterator &arg)
const
{
    return _ptr > arg._ptr;
}

template<typename T>
bool
basic_auto_array_iterator<T>::operator <=(const basic_auto_array_iterator &arg)
const
{
    return _ptr <= arg._ptr;
}

template<typename T>
bool
basic_auto_array_iterator<T>::operator >=(const basic_auto_array_iterator &arg)
const
{
    return _ptr >= arg._ptr;
}

template<typename T>
bool
basic_auto_array_iterator<T>::operator ==(const basic_auto_array_iterator &arg)
const
{
    return _ptr == arg._ptr;
}

template<typename T>
bool
basic_auto_array_iterator<T>::operator !=(const basic_auto_array_iterator &arg)
const
{
    return _ptr != arg._ptr;
}

template<typename T>
class basic_auto_array {
public:
    typedef basic_auto_array_iterator<const T> const_iterator;
    typedef basic_auto_array_iterator<T> iterator;
    typedef size_t size_type;
    basic_auto_array(size_type size); // constructor
    ~basic_auto_array(); // destructor
    template<typename U> T &operator [](const U &arg);
    template<typename U> const T &operator [](const U &arg) const;
    operator T *(); // type cast operator
    operator const T *() const; // type cast operator
    template<typename U> T *operator +(const U &arg);
    template<typename U> const T *operator +(const U &arg) const;
    iterator begin();
    const_iterator begin() const;
    iterator end(size_type size);
    const_iterator end(size_type size) const;
protected:
    T *_array;
    basic_auto_array(const basic_auto_array &arg); // copy constructor
    void release_array();
private:
    // no assignment operator
    basic_auto_array &operator =(const basic_auto_array &arg);
};

template<typename T>
basic_auto_array<T>::basic_auto_array(size_type size) {
    _array = new T[size];
}

template<typename T>
basic_auto_array<T>::basic_auto_array(const basic_auto_array &arg) {
    _array = arg._array;
}

template<typename T>
basic_auto_array<T>::~basic_auto_array() {
    _array = NULL;
}

template<typename T>
template<typename U>
T &basic_auto_array<T>::operator [](const U &arg) {
    assert(_array != NULL);
    return _array[arg];
}

template<typename T>
template<typename U>
const T &basic_auto_array<T>::operator [](const U &arg) const {
    assert(_array != NULL);
    return _array[arg];
}

template<typename T>
basic_auto_array<T>::operator T *() {
    return _array;
}

template<typename T>
basic_auto_array<T>::operator const T *() const {
    return _array;
}

template<typename T>
template<typename U>
T *basic_auto_array<T>::operator +(const U &arg) {
    return _array + arg;
}

template<typename T>
template<typename U>
const T *basic_auto_array<T>::operator +(const U &arg) const {
    return _array + arg;
}

template<typename T>
typename basic_auto_array<T>::iterator basic_auto_array<T>::begin() {
    return iterator(_array);
}

template<typename T>
typename basic_auto_array<T>::const_iterator basic_auto_array<T>::begin() const
{
    return const_iterator(_array);
}

template<typename T>
typename basic_auto_array<T>::iterator basic_auto_array<T>::end(size_type size)
{
    return iterator(_array + size);
}

template<typename T>
typename basic_auto_array<T>::const_iterator
basic_auto_array<T>::end(size_type size)
const
{
    return const_iterator(_array + size);
}

template<typename T>
void basic_auto_array<T>::release_array() {
    delete _array;
}

template<typename T>
class auto_array: public basic_auto_array<T> {
public:
    typedef size_t size_type;
    auto_array(size_type size);
    ~auto_array(); // destructor
private:
    auto_array(); // no default constructor
    auto_array(const auto_array &); // no copy constructor
    auto_array &operator =(const auto_array &arg); // no assignment operator
};

template<typename T>
auto_array<T>::auto_array(size_type size): basic_auto_array<T>(size) {
}

template<typename T>
auto_array<T>::~auto_array() {
    this->release_array();
}

template<typename T>
class smart_array: public basic_auto_array<T> {
public:
    typedef size_t size_type;
    smart_array(size_type size);
    smart_array(const smart_array &arg); // copy constructor
    ~smart_array(); // destructor
    smart_array &operator =(const smart_array &arg); // assignment operator
    static void swap(smart_array &arg1, smart_array &arg2);
private:
    mutable const smart_array *_pprev;
    mutable const smart_array *_pnext;
    smart_array(); // no default constructor
    void link(const smart_array *pprev);
};

template<typename T>
smart_array<T>::smart_array(size_type size): basic_auto_array<T>(size) {
    _pprev = this;
    _pnext = this;
}

template<typename T>
smart_array<T>::smart_array(const smart_array &arg): basic_auto_array<T>(arg) {
    link(&arg);
}

template<typename T>
smart_array<T>::~smart_array() {
    const smart_array *pprev = _pprev;
    if (pprev == this)
        this->release_array();
    else {
        const smart_array<T> *pnext = _pnext;
        pnext->_pprev = pprev;
        pprev->_pnext = pnext;
        _pnext = this;
        _pprev = this;
    }
}

template<typename T>
smart_array<T> &smart_array<T>::operator =(const smart_array &arg) {
    if (this != &arg) {
        const smart_array *pprev = _pprev;
        if (pprev == this)
            this->release_array();
        else {
            const smart_array *pnext = _pnext;
            pnext->_pprev = pprev;
            pprev->_pnext = pnext;
        }
        this->_array = arg._array;
        link(&arg);
    }
    return this;
}

template<typename T>
void smart_array<T>::link(const smart_array *pprev) {
    const smart_array *pnext = pprev->_pnext;
    (_pprev = pprev)->_pnext = this;
    (_pnext = pnext)->_pprev = this;
}

template<typename T>
void smart_array<T>::swap(smart_array &arg1, smart_array &arg2) {
    T *array1 = arg1._array;
    T *array2 = arg2._array;
    const smart_array *pprev1 = arg1._pprev;
    const smart_array *pprev2 = arg2._pprev;
    const smart_array *pnext1 = arg1._pnext;
    const smart_array *pnext2 = arg2._pnext;
    arg1._array = array2;
    arg2._array = array1;
    arg1._pprev = pprev2;
    arg2._pprev = pprev1;
    arg1._pnext = pnext2;
    arg2._pnext = pnext1;
    pprev1->_pnext = &arg2;
    pprev2->_pnext = &arg1;
    pnext1->_pprev = &arg2;
    pnext2->_pprev = &arg1;
}
